/*
 * Copyright (c) 2005-2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/types.h>

#include <errno.h>
#include <inttypes.h>
#include <math.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <errnoLib.h>
#include <portLib.h>

#include <mor/pscan.h>
#include <mor/harmonica.h>

#include "neoTime.h"
#include "utils.h"

#define NMTNodeID 0x00
#define HeartbeatTimeMS 1000
/**
 ** Interface with the Elmo 'harmonica' motor controller
 **/
/*----------------------------------------------------------------------*/
/*
 * Send a command to the byte interpreter
 */
STATUS
intprtSetInt(CAN_HARMONICA_STR *h, int dataLen, char cmdChar1, char cmdChar2,
    int index, int data)
{
	CAN_MSG m;

	m.id = h->canOpen->rxPDO2;
	m.len = dataLen;
	m.type = 0;
	m.data[0] = cmdChar1;
	m.data[1] = cmdChar2;

	m.data[2] = index & 0xff;
	m.data[3] = index >> 8;

	m.data[4] = data & 0x000000ff;
	m.data[5] = (data & 0x0000ff00) >> 8;
	m.data[6] = (data & 0x00ff0000) >> 16;
	m.data[7] = (data & 0xff000000) >> 24;
	/* printf("intprtSetInt: ");
	   canMsgShow(&m);*/
	return pscanTransmitMsg(h->dev, &m);
}

/*----------------------------------------------------------------------*/
static void
sendSDODownload(CAN_HARMONICA_STR *h, int objIndex, int subIndex, int data)
{
	CAN_MSG m;
	const int ciInitDownloadReq = 0x20;
	const int ciNrBytesNoData = 0x00;
	const int ciExpedited = 0x02;
	const int ciDataSizeInd = 0x01;

	m.len = 8;
	m.type = 0;
	m.id = h->canOpen->rxSDO;
	m.data[0] = ciInitDownloadReq | (ciNrBytesNoData << 2)
	    | ciExpedited | ciDataSizeInd;
	m.data[1] = (objIndex & 0x00ff);
	m.data[2] = (objIndex & 0xff00) >> 8;
	m.data[3] = subIndex;
	m.data[4] = (data & 0x000000ff);
	m.data[5] = (data & 0x0000ff00) >> 8;
	m.data[6] = (data & 0x00ff0000) >> 16;
	m.data[7] = (data & 0xff000000) >> 24;
	pscanTransmitMsg(h->dev, &m);
}

/*----------------------------------------------------------------------*/
static STATUS
harmonicaSetMotionType(CAN_HARMONICA_STR *h)
{
	int maxAcc, maxDcc;

	maxAcc = h->driveParam.accIncrS;
	maxDcc = h->driveParam.decIncrS;

	switch (h->motionType) {
	case MOTIONTYPE_POSCTRL:
		harmonicaStop(h);
		mssleep(20);
		/* switch unit-mode */
		intprtSetInt(h, 8, 'U', 'M', 0, 5);
		/* Set target radius to X increments */
		intprtSetInt(h, 8, 'T', 'R', 1, 15);
		/* Set target time to X ms */
		intprtSetInt(h, 8, 'T', 'R', 2, 200);

		/* Set maximum acceleration to X inc/s^2 */
		intprtSetInt(h, 8, 'A', 'C', 0, maxAcc);
		/* Set maximum deceleration to X inc/s^2 */
		intprtSetInt(h, 8, 'D', 'C', 0, maxDcc);
		break;

	case MOTIONTYPE_VELCTRL:
		harmonicaStop(h);
		/* switch unit mode */
		intprtSetInt(h, 8, 'U', 'M', 0, 2);
		/* set profiler mode */
		intprtSetInt(h, 8, 'P', 'M', 0, 1);

		/* Set maximum acceleration to X inc/s^2 */
		intprtSetInt(h, 8, 'A', 'C', 0, maxAcc);
		/* Set maximum deceleration to X inc/s^2 */
		intprtSetInt(h, 8, 'D', 'C', 0, maxDcc);
		break;
	default:
		return ERROR;
	}
	mssleep(20);
	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Request status from the motor contoller
 */
static void
harmonicaRequestStatus(CAN_HARMONICA_STR *h)
{
	intprtSetInt(h, 4, 'S', 'R', 0, 0);
}

void
harmonicaRequestCurrent(CAN_HARMONICA_STR *h)
{
	intprtSetInt(h, 4, 'I', 'Q', 0, 0);
}

void
harmonicaRequestDigIn(CAN_HARMONICA_STR *h)
{
	intprtSetInt(h, 4, 'I', 'P', 0, 0);
}

/*----------------------------------------------------------------------*/
/*
 * Compute a velocity estimation, based on the position difference
 */
static double
estimVel(CAN_HARMONICA_STR *h, double pos)
{
	double vel;
	double dt;

	clock_gettime(CLOCK_REALTIME, h->currentTime);
	dt = absDiffTime(h->currentTime, h->velCalcTime);
	vel = (pos - h->oldPos)/dt;
	h->oldPos = pos;
	clock_gettime(CLOCK_REALTIME, h->velCalcTime);
	return vel;
}

/*----------------------------------------------------------------------*/
/*
 * Interpret the contents of the controller's status register
 */
static STATUS
evalStatusRegister(CAN_HARMONICA_STR *h, int status)
{
	STATUS ret;

	if (isBitSet(status, 0)) {
		if (h->motorState != ST_MOTOR_FAILURE) {
			fprintf(stderr, "Drive %d error: ",
				h->driveParam.driveIdent);
			if ((status & 0x0000000e) == 2)
				fprintf(stderr,
				    " under voltage\n");
			if ((status & 0x0000000e) == 4)
				fprintf(stderr,
				    " over voltage\n");
			if ((status & 0x0000000e) == 10)
				fprintf(stderr,
				    " short circuit\n");
			if ((status & 0x0000000e) == 12)
				fprintf(stderr,
				    " overheating\n");
		}
		h->newMotorState = ST_MOTOR_FAILURE;
		ret = ERROR;
	} else if (isBitSet(status, 6)) {
		/* Motor failure */
		if (h->motorState != ST_MOTOR_FAILURE) {
			/* request detailed description of failure */
			intprtSetInt(h, 4, 'M', 'F', 0, 0);
			clock_gettime(CLOCK_REALTIME, h->failureStartTime);
		}
		h->newMotorState = ST_MOTOR_FAILURE;
		ret = ERROR;
	} else {
		/* no error */
		ret = OK;
		/* general status bits */
		if (isBitSet(status, 4)) {
			if (h->motorState != ST_OPERATION_ENABLED) {
				fprintf(stderr, "Motor %d operation enabled\n",
				    h->driveParam.driveIdent);
				clock_gettime(CLOCK_REALTIME,
				    h->failureStartTime);
			}
			h->newMotorState = ST_OPERATION_ENABLED;
		} else {
			if (h->motorState != ST_OPERATION_DISABLED) {
				fprintf(stderr,
				    "Motor %d operation disabled\n",
				    h->driveParam.driveIdent);
			}
			h->newMotorState = ST_OPERATION_DISABLED;
		}
		/* current limit */
		if (isBitSet(status, 13)) {
			if (!h->currentLimitOn) {
				fprintf(stderr, "Motor %d current limit on\n",
				    h->driveParam.driveIdent);
			}
			h->currentLimitOn = true;
		} else {
			h->currentLimitOn = false;
		}
	}
	/* change state */
	h->motorState = h->newMotorState;
	return ret;
}

/*----------------------------------------------------------------------*/
/*
 * Interprets motor failure error codes
 */
static void
evalMotorFailure(CAN_HARMONICA_STR *h, int failure)
{
	fprintf(stderr, "Motor %d: ", h->driveParam.driveIdent);
	if (isBitSet(failure, 2)) {
		fprintf(stderr, "feedback loss\n");
	} else if (isBitSet(failure, 3)) {
		fprintf(stderr, "peak current exceeded\n");
	} else 	if (isBitSet(failure, 7)) {
		fprintf(stderr, "speed track error\n");
	} else 	if (isBitSet(failure, 8)) {
		fprintf(stderr, "position track error\n");
	} else 	if (isBitSet(failure, 17)) {
		fprintf(stderr, "speed limit exceeded\n");
	} else if (isBitSet(failure, 21)) {
		fprintf(stderr, "motor stuck");
	} else if (isBitSet(failure, 12)) {
		switch ((failure & 0xe000) >> 13) {
		case 0:
			fprintf(stderr, "OK\n");
			break;
		case 1:
			fprintf(stderr, "under voltage\n");
			break;
		case 2:
			fprintf(stderr, "over voltage\n");
			break;
		case 5:
			fprintf(stderr, "short circuit\n");
			break;
		case 6:
			fprintf(stderr, "overheating\n");
			break;
		default:
			fprintf(stderr, "reserved error code ??\n");
			break;
		}
	} else {
		fprintf(stderr, "unknown failure 0x%x\n", failure);
	}
}

/*----------------------------------------------------------------------*/
/*
 * Interpret digital inputs
 */
static void
decodeDigIn(CAN_HARMONICA_STR *h)
{
	if (h->digIn & HARMONICA_INPUT_FORWARD_LIMIT)
		printf("Reached FLS\n");
	if (h->digIn & HARMONICA_INPUT_REVERSE_LIMIT)
		printf("Reached RLS\n");
}


/*----------------------------------------------------------------------*/
/*
 * Initialize communication with the motor controller
 */
CAN_HARMONICA_STR *
harmonicaInit(CAN_DEV *dev, int id,
    CAN_MOTION_TYPE motionType,
    const CAN_OPEN_PARAM_STR *param,
    const CAN_DRIVE_PARAMS *driveParam)
{
	CAN_HARMONICA_STR *h;

	h = (CAN_HARMONICA_STR *)malloc(sizeof(CAN_HARMONICA_STR));
	if (h == NULL) {
		errnoSet(S_harmonicaLib_NO_MEMORY);
		return NULL;
	}
	h->dev = dev;
	h->divForRequestStatus = 10;
	h->canTimeout = 6;
	h->statusCtrl = 0;
	h->posGearMeasRad = 0;
	h->prevPosGearMeasRad = 0;
	h->velGearMeasRadS = 0;
	h->motionType = motionType;

	h->watchdogActive = false;
	h->currentTime = (struct timespec *)malloc(sizeof(struct timespec));
	h->watchdogTime = (struct timespec *)malloc(sizeof(struct timespec));
	h->velCalcTime = (struct timespec *)malloc(sizeof(struct timespec));
	h->failureStartTime =
	    (struct timespec *)malloc(sizeof(struct timespec));
	h->sendTime = (struct timespec *)malloc(sizeof(struct timespec));

	clock_gettime(CLOCK_REALTIME, h->velCalcTime);

	h->digIn = 0;

	h->countRequestDiv = 0;

	h->motorState = ST_PRE_INITIALIZED;
	h->currentLimitOn = false;

	memcpy(&h->driveParam, driveParam, sizeof(struct CAN_DRIVE_PARAMS));
	h->driveParam.posGearRadToPosMotIncr = driveParam->encIncrPerRevMot
	    * driveParam->gearRatio * driveParam->beltRatio / (2 * M_PI);
	h->driveParam.driveIdent = id;

	h->canOpen = param;
	/* stop all emissions of TPD01 */
	sendSDODownload(h, 0x1a00, 0, 0);

	clock_gettime(CLOCK_REALTIME, h->sendTime);
	clock_gettime(CLOCK_REALTIME, h->currentTime);

	return h;
}

/*----------------------------------------------------------------------*/
/*
 * Ends the commnunication with a motor controller
 */
void
harmonicaEnd(CAN_HARMONICA_STR *h)
{
	harmonicaStop(h);

	free(h->currentTime);
	free(h->watchdogTime);
	free(h->velCalcTime);
	free(h->failureStartTime);
	free(h->sendTime);

	free(h);
}

/*----------------------------------------------------------------------*/
/*
 * request initialization of the motor controller
 */
STATUS
harmonicaInitCtrl(CAN_HARMONICA_STR *h)
{
	CAN_MSG msg;
	int incrRevWheel = h->driveParam.gearRatio * h->driveParam.beltRatio
	    * h->driveParam.encIncrPerRevMot * 3;
	int count,  posCount;
	int ret = OK;

	// printf("-- initCtrl %d\n", h->driveParam.driveIdent);
	h->motorState = ST_PRE_INITIALIZED;

	harmonicaStop(h);
	mssleep(20);
	intprtSetInt(h, 8, 'X', 'M', 2, incrRevWheel * 5000);
	mssleep(20);
	intprtSetInt(h, 8, 'X', 'M', 1, -incrRevWheel * 5000);
	mssleep(20);

	harmonicaSetMotionType(h);

	/* Set position counter to zero */
	intprtSetInt(h, 8, 'P', 'X', 0, 0);
	count = 0;
	while (1) {
		pscanReceiveMsg(h->dev, &msg);
		if (msg.data[0] == 'P' && msg.data[1] == 'X') {
			posCount = canMsgGetInt(&msg, 4);
			h->posGearMeasRad = h->driveParam.sign *
			    driveParamPosMotIncrToPosGearRad(&h->driveParam,
				posCount);
			h->prevPosGearMeasRad = h->posGearMeasRad;
			break;
		}
		if (count > 300) {
			printf("drive %d: initial position not set\n",
			    h->driveParam.driveIdent);
			ret = ERROR;
			break;
		}
		mssleep(10);
		count++;
	}

	/*
	 * Set PDO mapping :
	 * TPD01:
	 *  - position
	 *  - velocity
	 */
	/* stop all emissions of TPD01 */
	sendSDODownload(h, 0x1a00, 0, 0);

	/* position 4 byte of TPD01 */
	sendSDODownload(h, 0x1a00, 1, 0x60640020);

	/* velocity 4 byte of TPD01 */
	sendSDODownload(h, 0x1a00, 2, 0x60690020);

	/* transmission type "synch" */
	sendSDODownload(h, 0x1800, 2, 1);

	/* activate mapped objects */
	sendSDODownload(h, 0x1a00, 0, 2);

	return ret;
}

/*----------------------------------------------------------------------*/
/*
 * start the motor controller
 */
STATUS
harmonicaStart(CAN_HARMONICA_STR *h)
{
	CAN_MSG m;
	struct timespec ts;
	int cnt, rc;

	/* start motor */
	// printf("-- start motor %d\n", h->driveParam.driveIdent);
	intprtSetInt(h, 8, 'M', 'O', 0, 1);
	mssleep(20);
	/* flush CAN bus */
	while (pscanReceiveMsg(h->dev, &m) == OK)
		;

	harmonicaRequestStatus(h);

	/* wait a most 10s for motor to become enabled */
	cnt = 0;
	while (h->motorState != ST_OPERATION_ENABLED && cnt < 10) {
		pscanReceiveMsgWait(h->dev, &m, 1000000);
		if (m.data[0] == 'S' && m.data[1] == 'R') {
			h->statusCtrl = canMsgGetInt(&m, 4);
			evalStatusRegister(h, h->statusCtrl);
			break;
		}
		cnt++;
		clock_gettime(CLOCK_REALTIME, &ts);
		ts.tv_sec++;
		do {
			rc = clock_nanosleep(CLOCK_REALTIME,
			    TIMER_ABSTIME, &ts, NULL);
		} while (rc == -1 && errno == EINTR);
	} /* while */
	if (h->motorState != ST_OPERATION_ENABLED) {
		errnoSet(S_harmonicaLib_MOTOR_FAILURE);
		return ERROR;
	}
	clock_gettime(CLOCK_REALTIME, h->watchdogTime);
	clock_gettime(CLOCK_REALTIME, h->sendTime);
	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * shut down the motor controller
 */
void
harmonicaStop(CAN_HARMONICA_STR *h)
{
	/* Motor off */
	intprtSetInt(h, 8, 'M', 'O', 0, 0);
}

/*----------------------------------------------------------------------*/
/*
 * reset the motor controller
 */
STATUS
harmonicaReset(CAN_HARMONICA_STR *h)
{
	printf("Resetting drive %d\n", h->driveParam.driveIdent);
	pscanReset(h->dev);

	/* Init motor */
	if (harmonicaInitCtrl(h) == ERROR) {
		return ERROR;
	}
	/* Start motor */
	if (harmonicaStart(h) == ERROR) {
		return ERROR;
	}
	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Activate the hardware watchdog
 */
STATUS
harmonicaStartWatchdog(CAN_HARMONICA_STR *h, bool started)
{

	if (started) {
		h->watchdogActive = true;
		/* consumer (PC) heartbeat time */
		sendSDODownload(h, 0x1016, 1,
		    (NMTNodeID << 16) | HeartbeatTimeMS);
		/* error behaviour after failure */
		sendSDODownload(h, 0x1029, 1, 2);
		/* motor behaviour after heartbeat failure: "quick stop" */
		sendSDODownload(h, 0x6007, 0, 3);
		/* activate emergency events: "heartbeat event" */
		sendSDODownload(h, 0x2f21, 0, 0x8);
		mssleep(20);
	} else {
		/* Stop watchdog */
		h->watchdogActive = false;
		/* motor behaviour after heartbead failure: no action */
		sendSDODownload(h, 0x6007, 0, 0);
		/* de-activiate emergency events */
		sendSDODownload(h, 0x2f21, 0, 0x0);
		mssleep(20);
	}

	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Request a given position and velocity and ask for data updates
 */
void
harmonicaSetGearPosVelRadS(CAN_HARMONICA_STR *h,
    double posGearRad, double velGearRadS)
{
	int posEncIncr;
	int velEncIncrPeriod;
	CAN_DRIVE_PARAMS *p = &h->driveParam;
	CAN_MSG msg;
	double dt;

	driveParamPosVelRadToIncr(p, posGearRad, velGearRadS,
	    &posEncIncr, &velEncIncrPeriod);
	posEncIncr *= driveParamGetSign(p);
	if (velEncIncrPeriod > driveParamGetVelMax(p)) {
		velEncIncrPeriod = driveParamGetVelMax(p);
	}
	if (velEncIncrPeriod < -driveParamGetVelMax(p)) {
		velEncIncrPeriod = -driveParamGetVelMax(p);
	}
	h->refVelGearIncS = velEncIncrPeriod;
	if (h->motionType == MOTIONTYPE_POSCTRL) {
		intprtSetInt(h, 8, 'S', 'P', 0, velEncIncrPeriod);
		intprtSetInt(h, 8, 'P', 'A', 0, posEncIncr);
		intprtSetInt(h, 4, 'B', 'G', 0, 0);
	} else if (h->motionType == MOTIONTYPE_VELCTRL) {
		velEncIncrPeriod *= driveParamGetSign(p);
		intprtSetInt(h, 8, 'J', 'V', 0, velEncIncrPeriod);
		intprtSetInt(h, 4, 'B', 'G', 0, 0);
	}
	/* send heartbeat */
	msg.id = 0x700;
	msg.len = 5;
	msg.type = 0;
	memset(msg.data, 0, 8);
	pscanTransmitMsg(h->dev, &msg);

	/* request pos and vel by TPDO1, triggerd by SYNC msg */
	canMsgSet(&msg, 0x80, 0, 0, 0, 0, 0, 0, 0, 0);
	msg.len = 0;

	pscanTransmitMsg(h->dev, &msg);

	clock_gettime(CLOCK_REALTIME, h->currentTime);
	dt = absDiffTime(h->currentTime, h->sendTime);
	if (dt > 1.0) {
		fprintf(stderr,
		    "Time between send velocity of motor %d is too large %f\n",
		    driveParamGetDriveIdent(p), dt);
	}
	clock_gettime(CLOCK_REALTIME, h->sendTime);
	/* request status */
	h->countRequestDiv++;
	if (h->countRequestDiv > h->divForRequestStatus) {
		harmonicaRequestStatus(h);
		h->countRequestDiv = 0;
	}
}

/*----------------------------------------------------------------------*/
/*
 * request a given velocity, and ask for data updates
 * also ask for status every divForRequestStatus calls
 */
void
harmonicaSetGearVelRadS(CAN_HARMONICA_STR *h, double vel)
{
	int velEncIncrPeriod;
	CAN_DRIVE_PARAMS *p = &h->driveParam;
	CAN_MSG m;
	double dt;

	velEncIncrPeriod = p->sign
	    * driveParamVelGearRadSToVelMotIncrPeriod(p, vel);

	if (velEncIncrPeriod > driveParamGetVelMax(p))
		velEncIncrPeriod = driveParamGetVelMax(p);
	if (velEncIncrPeriod < -driveParamGetVelMax(p))
		velEncIncrPeriod = -driveParamGetVelMax(p);
	h->refVelGearIncS = velEncIncrPeriod;

	intprtSetInt(h, 8, 'J', 'V', 0, velEncIncrPeriod);
	intprtSetInt(h, 4, 'B', 'G', 0, 0);

	/* request pos and vel by TPDO1, triggered by SYNC msg */
	m.id = 0x80;
	m.len = 0;
	m.type = 0;
	memset(m.data, 0, 8);
	pscanTransmitMsg(h->dev, &m);

	/* send heartbeat */
	m.id = 0x700;
	m.len = 5;
	m.type = 0;
	memset(m.data, 0, 8);
	pscanTransmitMsg(h->dev, &m);

	clock_gettime(CLOCK_REALTIME, h->currentTime);
	dt = absDiffTime(h->currentTime, h->sendTime);
	if (h->watchdogActive && dt > 1.0) {
		fprintf(stderr,
		    "Time between send velocity of motor %d is too large %f\n",
		    driveParamGetDriveIdent(p), dt);
	}
	clock_gettime(CLOCK_REALTIME, h->sendTime);
	/* request status */
	h->countRequestDiv++;
	if (h->countRequestDiv > h->divForRequestStatus) {
		harmonicaRequestStatus(h);
		h->countRequestDiv = 0;
	}
}

/*----------------------------------------------------------------------*/
/*
 * request a given velocity, and ask for data updates
 * also ask for status every divForRequestStatus calls
 */
void
harmonicaSetGearVelIncr(CAN_HARMONICA_STR *h, int velEncIncrPeriod)
{
	CAN_DRIVE_PARAMS *p = &h->driveParam;
	CAN_MSG m;
	double dt;

	if (velEncIncrPeriod > driveParamGetVelMax(p)) {
		velEncIncrPeriod = driveParamGetVelMax(p);
		printf("limit %d > %d\n",
		    velEncIncrPeriod, driveParamGetVelMax(p));
	}
	if (velEncIncrPeriod < -driveParamGetVelMax(p)) {
		velEncIncrPeriod = -driveParamGetVelMax(p);
		printf("limit %d < %d\n",
		    velEncIncrPeriod, -driveParamGetVelMax(p));
	}
	h->refVelGearIncS = velEncIncrPeriod;

	intprtSetInt(h, 8, 'J', 'V', 0, velEncIncrPeriod);
	intprtSetInt(h, 4, 'B', 'G', 0, 0);

	/* request pos and vel by TPDO1, triggered by SYNC msg */
	m.id = 0x80;
	m.len = 0;
	m.type = 0;
	memset(m.data, 0, 8);
	pscanTransmitMsg(h->dev, &m);

	/* send heartbeat */
	m.id = 0x700;
	m.len = 5;
	m.type = 0;
	memset(m.data, 0, 8);
	pscanTransmitMsg(h->dev, &m);

	clock_gettime(CLOCK_REALTIME, h->currentTime);
	dt = absDiffTime(h->currentTime, h->sendTime);
	if (dt > 1.0) {
		fprintf(stderr,
		    "Time between send velocity of motor %d is too large %f\n",
		    driveParamGetDriveIdent(p), dt);
	}
	clock_gettime(CLOCK_REALTIME, h->sendTime);
	/* request status */
	h->countRequestDiv++;
	if (h->countRequestDiv > h->divForRequestStatus) {
		harmonicaRequestStatus(h);
		h->countRequestDiv = 0;
	}
}

/*----------------------------------------------------------------------*/

void
harmonicaGetPosRad(CAN_HARMONICA_STR *h, double *pPos)
{
	*pPos = h->posGearMeasRad;
}

void
harmonicaGetPosVelRadS(CAN_HARMONICA_STR *h, double *pPos, double *pVel)
{
	*pPos = h->posGearMeasRad;
	*pVel = h->velGearMeasRadS;
}

void
harmonicaGetDeltaPosVelRadS(CAN_HARMONICA_STR *h,
    double *pDeltaPos, double *pVel)
{
	*pDeltaPos = h->posGearMeasRad - h->prevPosGearMeasRad;
	*pVel = h->velGearMeasRadS;
	h->prevPosGearMeasRad = h->posGearMeasRad;
}

CAN_MOTOR_STATE
harmonicaGetMotorState(CAN_HARMONICA_STR *h)
{
	return h->motorState;
}

void
harmonicaGetCurrent(CAN_HARMONICA_STR *h, double *pCurr)
{
	*pCurr = h->motorCurr;
}

/*----------------------------------------------------------------------*/
/*
 * Display status of a motor controller on stdout
 */
void
harmonicaShow(CAN_HARMONICA_STR *h)
{
	printf("driveIdent: %d\n", driveParamGetDriveIdent(&h->driveParam));
	printf("motionType: %d\n", h->motionType);
	printf("statusCtrl: 0x%08x\n", h->statusCtrl);
	printf("ref velocity: %d (%f)\n", h->refVelGearIncS,
	       driveParamVelMotIncrPeriodToVelGearRadS(&h->driveParam,
						       h->refVelGearIncS));
	printf("velocity (rad/S): %f\n", h->velGearMeasRadS);
	printf("position (rad): %f\n", h->posGearMeasRad);
	switch (h->motorState) {
	case ST_PRE_INITIALIZED:
		printf("state: PRE_INITIALIZED\n");
		break;
	case ST_OPERATION_ENABLED:
		printf("state: OPERATION_ENABLED\n");
		break;
	case ST_OPERATION_DISABLED:
		printf("state: OPERATION_DISABLED\n");
		break;
	case ST_MOTOR_FAILURE:
		printf("state: MOTOR_FAILURE\n");
		break;
	default:
		printf("state: %d (unknown)\n", h->motorState);
	}
	printf("current limit: %s\n", h->currentLimitOn ? "On" : "Off");
	printf("------------------------------------------------------\n");
};

/*----------------------------------------------------------------------*/

/*
 * Decode CAN messages sent by the motor controller
 *
 */

bool
harmonicaDecode(CAN_HARMONICA_STR *h, CAN_MSG *msg)
{
	bool ret = false;
	int temp1, temp2, posCnt, failure, para;
	CAN_DRIVE_PARAMS *p = &h->driveParam;

	if (h == NULL)
		return false;

	if (msg->id == h->canOpen->txPDO1) {
		/* Position and velocity update */
		temp1 = canMsgGetInt(msg, 0);
		h->posGearMeasRad = driveParamGetSign(p)
		    * driveParamPosMotIncrToPosGearRad(p, temp1);
		temp2 = canMsgGetInt(msg, 4);
		h->velGearMeasRadS = driveParamGetSign(p)
		    * driveParamVelMotIncrPeriodToVelGearRadS(p, temp2);
		clock_gettime(CLOCK_REALTIME, h->watchdogTime);
		ret = true;
	}
	if (msg->id == h->canOpen->txPDO2) {
		if (msg->data[0] == 'P' && msg->data[1] == 'X') {
			/* current pos */
			posCnt = canMsgGetInt(msg, 4);
			h->posGearMeasRad = driveParamGetSign(p)
			    * driveParamPosMotIncrToPosGearRad(p, posCnt);
			h->velGearMeasRadS = estimVel(h, h->posGearMeasRad);
		} else if (msg->data[0] == 'P' && msg->data[1] == 'A') {
			/* position absolute */
		} else if (msg->data[0] == 'J' && msg->data[1] == 'V') {
			/* ?? */
		} else if (msg->data[0] == 'B' && msg->data[1] == 'G') {
			/* begin motion */
		} else if (msg->data[0] == 'U' && msg->data[1] == 'M') {
			/* user mode */
			para = canMsgGetInt(msg, 4);
			printf("-> UM: %d\n", para);
		} else if (msg->data[0] == 'I' && msg->data[1] == 'P') {
			/* digital in -> limit switches */
			h->digIn = canMsgGetInt(msg, 4) & 0x1fffff;
			// decodeDigIn(h);
		} else if (msg->data[0] == 'S' && msg->data[1] == 'P') {
			/* Speed for PTP mode */
			para = canMsgGetInt(msg, 4);
		} else if (msg->data[0] == 'S' && msg->data[1] == 'R') {
			/* status */
			h->statusCtrl = canMsgGetInt(msg, 4);
			evalStatusRegister(h, h->statusCtrl);
		} else if (msg->data[0] == 'M' && msg->data[1] == 'F') {
			/* motor failure */
			failure = canMsgGetInt(msg, 4);
			evalMotorFailure(h, failure);
		} else if (msg->data[0] == 'P' && msg->data[1] == 'M') {
			para = canMsgGetInt(msg, 4);
		} else if (msg->data[0] == 'A' && msg->data[1] == 'C') {
			para = canMsgGetInt(msg, 4);
		} else if (msg->data[0] == 'D' && msg->data[1] == 'C') {
			para = canMsgGetInt(msg, 4);
		} else if (msg->data[0] == 'H' && msg->data[1] == 'M') {
			/* homing status message */
			if (msg->data[4] == 0)
				h->homePos = true;
		} else if (msg->data[0] == 'I' && msg->data[1] == 'Q') {
			h->motorCurr = canMsgGetFloat(msg, 4);
		} else {
			fprintf(stderr, "harmonicaDecode: unknown msg %c%c\n",
			    msg->data[0], msg->data[1]);
		}
		clock_gettime(CLOCK_REALTIME, h->watchdogTime);
		ret = true;
	}
	if (msg->id == h->canOpen->txSDO) {
		posCnt = driveParamGetSign(p)*canMsgGetInt(msg, 4);
		h->posGearMeasRad =
		    driveParamPosMotIncrToPosGearRad(p, posCnt);
		h->velGearMeasRadS = estimVel(h, h->posGearMeasRad);
		clock_gettime(CLOCK_REALTIME, h->watchdogTime);
		ret = true;
	}
	return ret;
}

/*----------------------------------------------------------------------*/

/**
 ** Homing sequence for steering wheels (and spine)
 **/
void
harmonicaInitHoming(CAN_HARMONICA_STR *h)
{
	int posRef;

	posRef = h->driveParam.encOffsetIncr;
	/* disarm previous homing process */
	intprtSetInt(h, 8, 'H', 'M', 1, 0);
	mssleep(20);
	/* set the value to which the counter shall be resetted as soon as
	   the homing event occurs */
	intprtSetInt(h, 8, 'H', 'M', 2, posRef);
	mssleep(20);
	/* choose the chanel on which the controller listens for the homing
	   event */
	intprtSetInt(h, 8, 'H', 'M', 3, h->driveParam.homingDigIn);
	mssleep(20);
	/* choose the action that the controller shall perform after the
	   homing event occured '2' -> do noting */
	intprtSetInt(h, 8, 'H', 'M', 4, 2);
	mssleep(20);
	/* choose the setting of the position counter defined in step 2,
	   0 -> absolute setting of the position counter PX = HM[2] */
	intprtSetInt(h, 8, 'H', 'M', 5, 0);
	mssleep(20);
}

/*
 * Exec homing for one motor
 */
STATUS
harmonicaExecHoming(CAN_HARMONICA_STR *h)
{
	int count;
	CAN_MSG msg;

	int drive = driveParamGetDriveIdent(&h->driveParam);

	printf("Starting homing on drive %d\n", drive);

	/* make motor move */
	harmonicaSetGearVelIncr(h, h->driveParam.homingSpeed);

	/* arm the homing process */
	intprtSetInt(h, 8, 'H', 'M', 1, 1);

	/* flush messages */
	while (pscanReceiveMsg(h->dev, &msg) == OK)
		;

	count = 0;
	h->homePos = false;
	do {
		/* ask for status of homing process */
		if (intprtSetInt(h, 4, 'H', 'M', 1, 0) == ERROR) {
			printf("Error sending status request, drive %d\n",
			    drive);
			break;
		}
		/* Read and interpret messages */
		while (pscanReceiveMsgWait(h->dev, &msg, 0) != ERROR) {
			harmonicaDecode(h, &msg);
			if (h->homePos) {
				printf("Got homing signal on drive %d\n",
				    drive);
				break;
			}
		}
		if (h->homePos)	/* done */
			break;
		/* Refresh speed reference */
		harmonicaSetGearVelIncr(h, h->driveParam.homingSpeed);
		mssleep(10);	/* 10ms */
		count++;
	} while (count < (h->driveParam.homingDelay * 10000));

	/* Normally the elmo controller stopped the motor already
	 * if the homing position was reached */
	harmonicaSetGearVelIncr(h, 0);

	if (!h->homePos) {
		printf("Homing failed on drive %d\n", drive);
		return ERROR;
	}
	return OK;
}
