/*
 * Copyright (c) 2005-2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <fcntl.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <portLib.h>
#include <errnoLib.h>

#include "mor/pscan.h"

/* #define CAN_BUS_DEBUG 1 */

/**
 ** Interface with the Peak Systems CAN bus controller (Parallel port Dongle)
 **/
/*----------------------------------------------------------------------*/
/*
 * Initialize the communication with the CAN bus controller
 */
struct CAN_DEV *
pscanInit(char *device)
{
	CAN_DEV *dev;
	int ret;

	dev = (CAN_DEV *)calloc(1, sizeof(CAN_DEV));
	if (dev == NULL) {
		errnoSet(S_pscanLib_NO_MEMORY);
		return NULL;
	}
	dev->handle = LINUX_CAN_Open(device, O_RDWR);
	 if (dev->handle == 0) {
		errnoSet(S_pscanLib_OPEN_ERROR);
		free(dev);
		return NULL;
	}
	ret = CAN_Init(dev->handle, CAN_BAUD_1M, CAN_INIT_TYPE_ST);
	if (ret != 0) {
		errnoSet(S_pscanLib_INIT_ERROR);
		free(dev);
		return NULL;
	}
	dev->initialized = true;
	return dev;
}
/*----------------------------------------------------------------------*/
/*
 * End the communication with the CAN bus controller
 */
void
pscanEnd(CAN_DEV *dev)
{
	CAN_Close(dev->handle);
	free(dev);
}

/*----------------------------------------------------------------------*/
/*
 * Reset CAN Bus
 */
STATUS
pscanReset(CAN_DEV *dev)
{
	CAN_MSG msg;

	canMsgSet(&msg, 0, 1, 0, 0, 0, 0, 0, 0, 0);
	msg.len = 2;
	return pscanTransmitMsg(dev, &msg);
}

/*----------------------------------------------------------------------*/
/*
 * Send a message on the BUS
 */
STATUS
pscanTransmitMsg(CAN_DEV *dev, CAN_MSG *m)
{
	TPCANMsg tmsg;
	int ret;

	if (!dev->initialized) {
		errnoSet(S_pscanLib_NOT_INITIALIZED);
		return ERROR;
	}
#ifdef CAN_BUS_DEBUG
	printf("-> "); canMsgShow(m);
#endif
	memset(&tmsg, 0, sizeof(TPCANMsg));
	/* copy message */
	tmsg.ID = m->id;
	tmsg.MSGTYPE = m->type;
	tmsg.LEN = m->len;
	memcpy(&tmsg.DATA, m->data, 8);

	CAN_Write(dev->handle, &tmsg);
	ret = CAN_Status(dev->handle);
	if (ret < 0) {
		errnoSet(S_pscanLib_WRITE_ERROR);
		return ERROR;
	}
	return OK;
}
/*----------------------------------------------------------------------*/
/*
 * Wait for a message and receive it
 *
 * usec: number of micro-seconds to wait : -1 => wait forever
 */
STATUS
pscanReceiveMsgWait(CAN_DEV *dev, CAN_MSG *m, int usec)
{
	TPCANRdMsg trmsg;
	int ret;

	if (!dev->initialized)
		return ERROR;

	trmsg.Msg.LEN = 8;
	trmsg.Msg.MSGTYPE = 0;
	trmsg.Msg.ID = 0;
	ret = LINUX_CAN_Read_Timeout(dev->handle, &trmsg, usec);
	if (ret != CAN_ERR_OK) {
		memset(m, 0, sizeof(CAN_MSG));
		errnoSet(S_pscanLib_READ_ERROR);
		return ERROR;
	}
	m->id = trmsg.Msg.ID;
	m->type = trmsg.Msg.MSGTYPE;
	m->len = trmsg.Msg.LEN;
	memcpy(m->data, &trmsg.Msg.DATA, 8);
#ifdef CAN_BUS_DEBUG
	printf("<- "); canMsgShow(m);
#endif
	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Non-Blocking check for next message and receive it
 */
STATUS
pscanReceiveMsg(CAN_DEV *dev, CAN_MSG *m)
{
	return pscanReceiveMsgWait(dev, m, 0);
}


/*----------------------------------------------------------------------*/
/*
 * Initialize a CAN message
 */
void
canMsgSet(CAN_MSG *m, int id,
    int p1, int p2, int p3, int p4, int p5, int p6, int p7, int p8)
{
	m->id = id;
	m->len = 8;
	m->type = 0;
	m->data[0] = p1;
	m->data[1] = p2;
	m->data[2] = p3;
	m->data[3] = p4;
	m->data[4] = p5;
	m->data[5] = p6;
	m->data[6] = p7;
	m->data[7] = p8;
}

/*----------------------------------------------------------------------*/
/*
 * Extract the "command" part of a CAN message (Neobotix devices only)
 */
int
canGetCmd(CAN_MSG *m)
{
	return (int)(m->data[7] >> 2);
}
/*----------------------------------------------------------------------*/
/*
 * Extract the status part of a CAN message (Neobotix devices only)
 */
int
canGetStatus(CAN_MSG *m)
{
	return (int)(m->data[7] & 0x3);
}

/*----------------------------------------------------------------------*/
/*
 * Extract a little-endian 32 bit integer from a CAN message at given offset
 */
int
canMsgGetInt(CAN_MSG *m, int offset)
{
	return (m->data[offset+3] << 24) | (m->data[offset+2] << 16)
	    | (m->data[offset+1] << 8) | m->data[offset];
}
/*
 * Extract a float from a CAN message at given offset
 */
float
canMsgGetFloat(CAN_MSG *m, int offset)
{
	union {
		int i;
		float f;
	} msg;

	msg.i = (m->data[offset+3] << 24) | (m->data[offset+2] << 16)
	    | (m->data[offset+1] << 8) | m->data[offset];
	return msg.f;
}

/*----------------------------------------------------------------------*/
/*
 * Display the raw contents of a CAN message on stdout
 */
void
canMsgShow(CAN_MSG *m)
{
	int i;

	printf("ID: 0x%x ", m->id);
	printf("Type: %d ", m->type);
	printf("len: %d ", m->len);
	printf("data:");
	for (i = 0; i < 8; i++) {
		printf(" 0x%02x", m->data[i]);
	}
	printf("\n");
}
